! ESA (C) 2000-2018 
!
! This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
!
!    SAS is free software: you can redistribute it and/or modify
!    it under the terms of the GNU General Public License as published by
!    the Free Software Foundation, either version 3 of the License, or
!    (at your option) any later version.
!
!    SAS is distributed in the hope that it will be useful,
!    but WITHOUT ANY WARRANTY; without even the implied warranty of
!    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!    GNU General Public License for more details.
!
!    You should have received a copy of the GNU General Public License
!    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
module recover16bits_mod
!  VNY 20.05.2004 : This module is aimed at restoring 
!  16 bits (32767 counts) in some of the pixels belonging to
!  bright sources (usually brighter than the magitude B=14).
!  The reason by which these counts are missed in not yet known.
!
   use dal
   use sasdal
   use errorHandling
!   use OmAllocation_mod      ! For memory allocation
!   use OmSourceparam_mod

   implicit none

   private

   public recover16bits

contains

subroutine recover16bits( &
    nx,                     &  ! The x-extent of the images 
    ny,                     &  ! The y-extent of the images 
    image,                   &  ! The image (including background) 
    photometryRadius,        &  ! The aperture radius (pixels) to be used for doing photometry
    nSources,               &  ! The number of sources
    sourceXCoord,            &  ! Array of x-coordinates of the sources
    sourceYCoord)               ! Array of y-coordinated of the source 

    ! Input and unmodified variables
    integer(int32), intent(in) :: nx, ny                      
    real(single), pointer              :: image(:, :)      
    real(single), intent(in) :: photometryRadius       
    integer (int32), intent(in)                      :: nSources
    real(kind=single), dimension(:), pointer :: sourceXCoord
    real(kind=single), dimension(:), pointer :: sourceYCoord

    ! The following structure is defined in  "omdetectsource_mod.f90"
    ! It stores all the parameters associated with a source 
    !type(SourceParameters), pointer :: source(:)       
 
    ! Local variables
    integer(int32) :: i, j, k, iSource, i1, j1 
    integer(int32) :: ix, iy, dxy, nHoles
    real(kind=single) :: sCounts, sPixels,  maxCounts
    logical :: flag
    integer(int32) :: xHole(1:10), yHole(1:10)

    if (nSources .lt. 1) return ! Return if no sources
    write(errstr,'(a)') "Recovering possible 16-bits losses in bright sources"
    call message(AppMsg, Verbose)
    dxy=photometryRadius/2
    write(*,*) " L51 nx=", nx, " ny=", ny
    !write(*,*) "******************************************************************"
    ! Find four sources with known coordinates
    !do iSource=0, nSources-1
    !   ix=sourceXCoord(iSource)
    !   iy=sourceYCoord(iSource)
    !   if (abs(ix-261) .lt. 2 .and. abs(iy-654) .lt. 2 &
    !        .or. abs(ix-133) .lt. 2 .and. abs(iy-666) .lt. 2 &
    !        .or. abs(ix-135) .lt. 2 .and. abs(iy-572) .lt. 2 &
    !        .or. abs(ix-811) .lt. 2 .and. abs(iy-612) .lt. 2) then
    !      write(*,'(i6,a,i5,a,i5)') &
    !           iSource, " X=", ix, " Y=", iy
    !      !sCounts=0.0
    !      !sPixels=0.0
    !      nHoles=0
    !      do i=ix-dxy+1,ix+dxy-1
    !         do j=iy-dxy+1,iy+dxy-1
    !            flag=.false.
    !            if (i-1 .ge. 0 .and. i+1 .lt. nx .and. j-1 .ge. 0 .and. j+1 .lt. ny) then
    !               if (image(i-1,j) .gt. image(i,j)+1000.0 .and. image(i+1,j) .gt. image(i,j)+1000.0) flag=.true.
    !               if (image(i,j-1) .gt. image(i,j)+1000.0 .and. image(i,j+1) .gt. image(i,j)+1000.0) flag=.true.
    !            endif
    !            if (flag) then
    !               write(*,*) " iHole=",i, " jHole=", j
    !               nHoles=nHoles+1
    !            endif
    !         enddo ! j
    !      enddo ! i
    !      !if (nHoles .gt. 0) then
    !      write(*,*) " nHoles=", nHoles
    !      !endif
    !   endif ! (abs(ix-261 
    !enddo  ! iSource
    write(*,*) "******************************************************************"
    k=0
    do iSource=0, nSources-1
       ix=sourceXCoord(iSource)
       iy=sourceYCoord(iSource)
       ! Compute the number of counts within the source
       sCounts=0.0
       sPixels=0.0
       maxCounts=0.0
       do i=ix-dxy,ix+dxy
          do j=iy-dxy,iy+dxy
             if (i .ge. 0 .and. i .lt. nx .and. j .ge. 0 .and. j .lt. ny) then
                sCounts=sCounts+image(i,j)
                sPixels=sPixels+1.0
                if (image(i,j) .gt. maxCounts) maxCounts=image(i,j)
             endif
          enddo ! j
       enddo ! i
       if (sPixels .gt. 0) then
          sCounts=sCounts/sPixels
       else
          sCounts=0.0
       endif
       if (sCounts .gt. 3000.0) then
          ! search for a hole
          nHoles=0
          xHole=-1
          yHole=-1
          do i=ix-dxy+1,ix+dxy-1
             do j=iy-dxy+1,iy+dxy-1
                flag=.false.
                if (i-1 .ge. 0 .and. i+1 .lt. nx .and. j-1 .ge. 0 .and. j+1 .lt. ny) then
                   if (image(i-1,j) .gt. image(i,j)+1000.0 .and. image(i+1,j) .gt. image(i,j)+1000.0) flag=.true.
                   if (image(i,j-1) .gt. image(i,j)+1000.0 .and. image(i,j+1) .gt. image(i,j)+1000.0) flag=.true.
                endif
                if (flag) then
                   k=k+1
                   i1=i
                   j1=j
                   nHoles=nHoles+1
                   if (nHoles .le. 10) then
                      xHole(nHoles)=i
                      yHole(nHoles)=j
                   endif ! nHoles
                   image(i,j)=image(i,j)+32767.0
                endif ! flag
             enddo ! j
          enddo ! i
          if (nHoles .gt. 0) then
             write(*,'(i6,i6,a,i5,a,i5, a,f7.1,a, f7.1, a, i3, a, i4, a, i4)') &
                  k, iSource, " X=", ix, " Y=", iy, " sCounts=", sCounts, &
                  " maxCounts=", maxCounts, " nHoles=", nHoles, " i1=", i1, " j1=", j1
          endif ! nHoles
       endif ! sCounts
    enddo  ! iSource
    write(*,*) " ---------- "

end subroutine recover16bits

end module recover16bits_mod

